import pytest
from MDAnalysis import SelectionError
from numpy.testing import assert_array_equal
from rdkit import Chem

from prolif.datafiles import datapath
from prolif.molecule import Molecule, mol2_supplier, pdbqt_supplier, sdf_supplier
from prolif.residue import ResidueId


class TestMolecule(pytest.BaseTestMixinRDKitMol):
    @pytest.fixture(scope="class")
    def mol(self, rdkit_mol):
        return Molecule(rdkit_mol)

    def test_mapindex(self, mol):
        for atom in mol.GetAtoms():
            assert atom.GetUnsignedProp("mapindex") == atom.GetIdx()

    def test_from_mda(self, u, ligand_rdkit):
        rdkit_mol = Molecule(ligand_rdkit)
        mda_mol = Molecule.from_mda(u, "resname LIG")
        assert rdkit_mol[0].resid == mda_mol[0].resid
        assert rdkit_mol.HasSubstructMatch(mda_mol) and mda_mol.HasSubstructMatch(
            rdkit_mol
        )

    def test_from_mda_empty_ag(self, u):
        ag = u.select_atoms("resname FOO")
        with pytest.raises(SelectionError, match="AtomGroup is empty"):
            Molecule.from_mda(ag)

    def test_from_rdkit(self, ligand_rdkit):
        rdkit_mol = Molecule(ligand_rdkit)
        newmol = Molecule.from_rdkit(ligand_rdkit)
        assert rdkit_mol[0].resid == newmol[0].resid

    def test_from_rdkit_default_resid(self):
        mol = Chem.MolFromSmiles("CCO")
        newmol = Molecule.from_rdkit(mol)
        assert newmol[0].resid == ResidueId("UNL", 1)

    def test_from_rdkit_resid_args(self):
        mol = Chem.MolFromSmiles("CCO")
        newmol = Molecule.from_rdkit(mol, "FOO", 42, "A")
        assert newmol[0].resid == ResidueId("FOO", 42, "A")

    @pytest.mark.parametrize("key", [0, 42, -1, "LYS49.A", ResidueId("LYS", 49, "A")])
    def test_getitem(self, mol, key):
        assert mol[key].resid is mol.residues[key].resid

    def test_iter(self, mol):
        for i, r in enumerate(mol):
            assert r.resid == mol[i].resid

    def test_n_residues(self, mol):
        assert mol.n_residues == mol.residues.n_residues


class SupplierBase:
    resid = ResidueId("UNL", 1, "")

    def test_len(self, suppl):
        assert len(suppl) == 9

    def test_returns_mol(self, suppl):
        mol = next(iter(suppl))
        assert isinstance(mol, Molecule)

    def test_monomer_info(self, suppl):
        mol = next(iter(suppl))
        resid = ResidueId.from_atom(mol.GetAtomWithIdx(0))
        assert resid == self.resid

    def test_index(self, suppl):
        mols = list(suppl)
        for index in [0, 2, 8, -1]:
            mol_i = suppl[index]
            assert_array_equal(mols[index].xyz, mol_i.xyz)


class TestPDBQTSupplier(SupplierBase):
    resid = ResidueId("LIG", 1, "G")

    @pytest.fixture
    def suppl(self):
        path = datapath / "vina"
        pdbqts = sorted(path.glob("*.pdbqt"))
        template = Chem.MolFromSmiles(
            "C[NH+]1CC(C(=O)NC2(C)OC3(O)C4CCCN4C(=O)"
            "C(Cc4ccccc4)N3C2=O)C=C2c3cccc4[nH]cc"
            "(c34)CC21"
        )
        return pdbqt_supplier(pdbqts, template)

    def test_pdbqt_hydrogens_stay_in_mol(self, ligand_rdkit):
        template = Chem.RemoveHs(ligand_rdkit)
        indices = []
        rwmol = Chem.RWMol(ligand_rdkit)
        rwmol.BeginBatchEdit()
        for atom in rwmol.GetAtoms():
            idx = atom.GetIdx()
            atom.SetIntProp("_MDAnalysis_index", idx)
            if atom.GetAtomicNum() == 1:
                if idx % 2:
                    indices.append(idx)
                else:
                    neighbor = atom.GetNeighbors()[0]
                    rwmol.RemoveBond(idx, neighbor.GetIdx())
                    rwmol.RemoveAtom(idx)
                    neighbor.SetNumExplicitHs(1)
        rwmol.CommitBatchEdit()
        pdbqt_mol = rwmol.GetMol()
        mol = pdbqt_supplier._adjust_hydrogens(template, pdbqt_mol)
        hydrogens = [
            idx
            for atom in mol.GetAtoms()
            if atom.HasProp("_MDAnalysis_index")
            and (idx := atom.GetIntProp("_MDAnalysis_index")) in indices
        ]
        assert hydrogens == indices


class TestSDFSupplier(SupplierBase):
    @pytest.fixture
    def suppl(self):
        path = str(datapath / "vina" / "vina_output.sdf")
        return sdf_supplier(path)


class TestMOL2Supplier(SupplierBase):
    @pytest.fixture
    def suppl(self):
        path = str(datapath / "vina" / "vina_output.mol2")
        return mol2_supplier(path)

    def test_mol2_starting_with_comment(self):
        path = str(datapath / "mol_comment.mol2")
        suppl = mol2_supplier(path)
        mol = next(iter(suppl))
        assert mol is not None



"""
Pickling support --- :mod:`prolif.pickling`
===========================================

This module provides the :class:`RDKitPickleHandler` class and a default instance
named `PICKLE_HANDLER` preconfigured to pickle all properties set on a molecule.

Note that on Windows the preconfigured handler will modify RDKit's default pickling
behavior upon importing the `prolif` package to ensure that the parallelization code
works as on other platforms.
"""
import sys
from typing import Optional

from rdkit import Chem

PROLIF_PICKLE_OPTIONS = (
    Chem.PropertyPickleOptions.AtomProps ^ Chem.PropertyPickleOptions.MolProps
)


class RDKitPickleHandler:
    def __init__(self, pickle_options: int) -> None:
        self.pickle_options = pickle_options
        self.default_pickle = self.get()
        self.is_patched = False

    @staticmethod
    def get() -> int:
        """Get RDKit's current pickle properties option"""
        return Chem.GetDefaultPickleProperties()

    def set(self, pickle_options: Optional[int] = None) -> None:
        """Set RDKit to the specified pickle options (or the one specified
        upon initializing if ``None``).
        """
        if pickle_options is None:
            pickle_options = self.pickle_options
        Chem.SetDefaultPickleProperties(pickle_options)

    def reset(self, force: bool = False) -> None:
        """Reset RDKit's pickle options to their default (before instantiating this
        class).
        """
        if force or not self.is_patched:
            self.set(self.default_pickle)

    def patch(self) -> None:
        """Patches property pickling on Windows.

        Notes
        -----
        For some reason RDKit properties pickling need to be set early (before the
        context manager of the parallel classes) to work as expected on Windows.
        """
        if sys.platform == "win32":
            self.set()
            self.is_patched = True


PICKLE_HANDLER = RDKitPickleHandler(PROLIF_PICKLE_OPTIONS)
PICKLE_HANDLER.patch()



import pytest
from PIL.PngImagePlugin import PngImageFile

import prolif as plf


@pytest.mark.parametrize(
    "kwargs, expected_size",
    [
        ({"residues_slice": slice(10)}, "width='800px' height='420px'"),
        ({"residues_slice": slice(10, 20)}, "width='800px' height='420px'"),
        ({"residues_slice": slice(0, 20, 2)}, "width='800px' height='420px'"),
        ({"size": (100, 100)}, "width='400px' height='1400px'"),
        ({"mols_per_row": 2}, "width='400px' height='3920px'"),
    ],
)
def test_display_residues_svg(
    protein_mol: plf.Molecule, kwargs: dict, expected_size: str
) -> None:
    img = plf.display_residues(protein_mol, **kwargs)
    assert "<svg" in img
    assert expected_size in img


def test_display_residues_png(protein_mol: plf.Molecule) -> None:
    img = plf.display_residues(protein_mol, slice(10), use_svg=False)
    assert isinstance(img, PngImageFile)
    assert img.size == (800, 420)



import numpy as np
import pytest
from MDAnalysis import Universe
from MDAnalysis.topology.guessers import guess_atom_element
from numpy.testing import assert_array_almost_equal
from rdkit import Chem
from rdkit.Chem.rdMolTransforms import ComputeCentroid

from prolif.datafiles import TOP, TRAJ, datapath
from prolif.molecule import Molecule, sdf_supplier


def pytest_sessionstart(session):
    if not datapath.exists():
        pytest.exit(
            f"Example data files are not accessible: {datapath!s} does not exist"
        )
    vina_path = datapath / "vina"
    if not vina_path.exists():
        pytest.exit(
            f"Example Vina data files are not accessible: {vina_path!s} does not exist"
        )
    # ugly patch to add Mixin class as attribute to pytest so that we don't have to
    # worry about relative imports in the test codebase
    setattr(pytest, "BaseTestMixinRDKitMol", BaseTestMixinRDKitMol)


@pytest.fixture(scope="session")
def u():
    return Universe(TOP, TRAJ)


@pytest.fixture(scope="session")
def rdkit_mol():
    return Chem.MolFromPDBFile(TOP, removeHs=False)


@pytest.fixture(scope="session")
def ligand_ag(u):
    return u.select_atoms("resname LIG")


@pytest.fixture(scope="session")
def ligand_rdkit(ligand_ag):
    return ligand_ag.convert_to.rdkit()


@pytest.fixture(scope="session")
def ligand_mol(ligand_ag):
    return Molecule.from_mda(ligand_ag)


@pytest.fixture(scope="session")
def protein_ag(u, ligand_ag):
    return u.select_atoms("protein and byres around 6.5 group ligand", ligand=ligand_ag)


@pytest.fixture(scope="session")
def protein_rdkit(protein_ag):
    return protein_ag.convert_to.rdkit()


@pytest.fixture(scope="session")
def protein_mol(protein_ag):
    return Molecule.from_mda(protein_ag)


@pytest.fixture(scope="session")
def sdf_suppl():
    path = str(datapath / "vina" / "vina_output.sdf")
    return sdf_supplier(path)


def from_mol2(f):
    path = str(datapath / f)
    u = Universe(path)
    elements = [guess_atom_element(n) for n in u.atoms.names]
    u.add_TopologyAttr("elements", np.array(elements, dtype=object))
    u.atoms.types = np.array([x.upper() for x in u.atoms.types], dtype=object)
    return Molecule.from_mda(u, force=True)


@pytest.fixture(scope="session")
def benzene():
    return from_mol2("benzene.mol2")


@pytest.fixture(scope="session")
def cation():
    return from_mol2("cation.mol2")


@pytest.fixture(scope="session")
def cation_false():
    return from_mol2("cation_false.mol2")


@pytest.fixture(scope="session")
def anion():
    return from_mol2("anion.mol2")


@pytest.fixture(scope="session")
def ftf():
    return from_mol2("facetoface.mol2")


@pytest.fixture(scope="session")
def etf():
    return from_mol2("edgetoface.mol2")


@pytest.fixture(scope="session")
def chlorine():
    return from_mol2("chlorine.mol2")


@pytest.fixture(scope="session")
def bromine():
    return from_mol2("bromine.mol2")


@pytest.fixture(scope="session")
def hb_donor():
    return from_mol2("donor.mol2")


@pytest.fixture(scope="session")
def hb_acceptor():
    return from_mol2("acceptor.mol2")


@pytest.fixture(scope="session")
def hb_acceptor_false():
    return from_mol2("acceptor_false.mol2")


@pytest.fixture(scope="session")
def xb_donor():
    return from_mol2("xbond_donor.mol2")


@pytest.fixture(scope="session")
def xb_acceptor():
    return from_mol2("xbond_acceptor.mol2")


@pytest.fixture(scope="session")
def xb_acceptor_false_xar():
    return from_mol2("xbond_acceptor_false_xar.mol2")


@pytest.fixture(scope="session")
def xb_acceptor_false_axd():
    return from_mol2("xbond_acceptor_false_axd.mol2")


@pytest.fixture(scope="session")
def ligand():
    return from_mol2("ligand.mol2")


@pytest.fixture(scope="session")
def metal():
    return from_mol2("metal.mol2")


@pytest.fixture(scope="session")
def metal_false():
    return from_mol2("metal_false.mol2")


class BaseTestMixinRDKitMol:
    def test_init(self, mol):
        assert isinstance(mol, Chem.Mol)

    def test_centroid(self, mol):
        expected = ComputeCentroid(mol.GetConformer())
        assert_array_almost_equal(mol.centroid, expected)

    def test_xyz(self, mol):
        expected = mol.GetConformer().GetPositions()
        assert_array_almost_equal(mol.xyz, expected)



"""
Plot interactions as a barcode --- :mod:`prolif.plotting.barcode`
=================================================================

.. versionadded:: 2.0.0

.. autoclass:: Barcode
   :members:

"""
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Dict, List, Literal, Optional, Tuple

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
from matplotlib.patches import Patch

from prolif.exceptions import RunRequiredError
from prolif.plotting.utils import IS_NOTEBOOK, separated_interaction_colors

if TYPE_CHECKING:
    from prolif.fingerprint import Fingerprint


class Barcode:
    """Creates a barcode plot of interactions.

    Parameters
    ----------
    df: pandas.DataFrame
        The DataFrame as obtained from :meth:`Fingerprint.to_dataframe()`

    Attributes
    ----------
    COLORS : dict
        Dictionnary of colors used in the plot for interactions.

    """

    COLORS: ClassVar[Dict[Optional[str], str]] = {
        None: "white",
        **separated_interaction_colors,
    }

    def __init__(self, df: pd.DataFrame) -> None:
        # mapping interaction type (HBond...etc.) to an arbitrary value which
        # corresponds to a color
        self.color_mapper = {
            interaction: value for value, interaction in enumerate(self.COLORS)
        }
        # reverse: map value to interaction type
        self.inv_color_mapper = {
            value: interaction for interaction, value in self.color_mapper.items()
        }
        # matplotlib colormap
        self.cmap = ListedColormap(list(self.COLORS.values()))

        # drop ligand level if single residue
        # else concatenate ligand with protein and drop ligand if peptide
        n_ligand_residues = len(np.unique(df.columns.get_level_values("ligand")))
        if n_ligand_residues == 1:
            df = df.droplevel("ligand", axis=1)
        else:
            df.columns = pd.MultiIndex.from_tuples(
                [(f"{items[0]}-{items[1]}", items[2]) for items in df.columns],
                names=["protein", "interaction"],
            )

        def _bit_to_color_value(s: pd.Series) -> pd.Series:
            """Replaces a bit value with it's corresponding color value"""
            interaction = s.name[-1]
            return s.apply(
                lambda v: self.color_mapper[interaction]
                if v
                else self.color_mapper[None]
            )

        self.df = df.astype(np.uint8).T.apply(_bit_to_color_value, axis=1)

    @classmethod
    def from_fingerprint(cls, fp: Fingerprint) -> Barcode:
        """Creates a barcode object from a fingerprint."""
        if not hasattr(fp, "ifp"):
            raise RunRequiredError(
                "Please run the fingerprint analysis before attempting to display results."
            )
        return cls(fp.to_dataframe())

    def display(
        self,
        figsize: Tuple[int, int] = (8, 10),
        dpi: int = 100,
        interactive: bool = IS_NOTEBOOK,
        n_frame_ticks: int = 10,
        residues_tick_location: Literal["top", "bottom"] = "top",
        xlabel: str = "Frame",
        subplots_kwargs: Optional[dict] = None,
        tight_layout_kwargs: Optional[dict] = None,
    ):
        """Generate and display the barcode plot.

        Parameters
        ----------
        figsize: Tuple[int, int] = (8, 10)
            Size of the matplotlib figure.
        dpi: int = 100
            DPI used for the matplotlib figure.
        interactive: bool
            Add hover interactivity to the plot (only relevant for notebooks). You may
            need to add ``%matplotlib notebook`` or ``%matplotlib ipympl`` for it to
            work as expected.
        n_frame_ticks: int = 10
            Number of ticks on the X axis. May use ±1 tick to have them evenly spaced.
        residues_tick_location: Literal["top", "bottom"] = "top"
            Whether the Y ticks appear at the top or at the bottom of the series of
            interactions of each residue.
        xlabel: str = "Frame"
            Label displayed for the X axis.
        subplots_kwargs: Optional[dict] = None
            Other parameters passed to :func:`matplotlib.pyplot.subplots`.
        tight_layout_kwargs: Optional[dict] = None
            Other parameters passed to :meth:`matplotlib.figure.Figure.tight_layout`.
        """
        if subplots_kwargs is None:
            subplots_kwargs = {}
        subplots_kwargs.setdefault("figsize", figsize)
        subplots_kwargs.setdefault("dpi", dpi)

        if tight_layout_kwargs is None:
            tight_layout_kwargs = {}
        tight_layout_kwargs.setdefault("pad", 1.2)

        # Plot as image
        fig, ax = plt.subplots(**subplots_kwargs)
        ax: plt.Axes
        im = ax.imshow(
            self.df.values,
            aspect="auto",
            cmap=self.cmap,
            vmin=0,
            vmax=max(self.color_mapper.values()),
        )

        # Frame ticks
        frames = self.df.columns
        max_ticks = len(frames) - 1
        # try to have evenly spaced ticks
        for effective_n_ticks in (n_frame_ticks, n_frame_ticks - 1, n_frame_ticks + 1):
            samples, step = np.linspace(0, max_ticks, effective_n_ticks, retstep=True)
            if step.is_integer():
                break
        else:
            samples = np.linspace(0, max_ticks, n_frame_ticks)
        indices = np.round(samples).astype(int)
        ax.xaxis.set_ticks(indices, frames[indices])
        ax.set_xlabel(xlabel)

        # Residues ticks
        n_items = len(self.df.index)
        residues = self.df.index.get_level_values("protein")
        interactions = self.df.index.get_level_values("interaction")
        if residues_tick_location == "top":
            indices = [
                i
                for i in range(n_items)
                if (i - 1 > 0 and residues[i - 1] != residues[i]) or i == 0
            ]
        else:
            indices = [
                i
                for i in range(n_items)
                if (i + 1 < n_items and residues[i + 1] != residues[i])
                or i + 1 == n_items
            ]
        ax.yaxis.set_ticks(indices, residues[indices])

        # legend
        values: List[int] = np.unique(self.df.values).tolist()
        try:
            values.pop(values.index(0))  # remove None color
        except ValueError:
            # 0 not in values (e.g. plotting a single frame)
            pass
        legend_colors = {
            self.inv_color_mapper[value]: im.cmap(value) for value in values
        }
        patches = [
            Patch(color=color, label=interaction)
            for interaction, color in legend_colors.items()
        ]
        ax.legend(handles=patches, bbox_to_anchor=(1.01, 1), loc=2, borderaxespad=0)

        # interactive
        if interactive:
            self._add_interaction_callback(
                fig,
                ax,
                im=im,
                frames=frames,
                residues=residues,
                interactions=interactions,
            )

        fig.tight_layout(**tight_layout_kwargs)
        return ax

    def _add_interaction_callback(self, fig, ax, *, im, frames, residues, interactions):
        annot = ax.annotate(
            "",
            xy=(0, 0),
            xytext=(5, 5),
            textcoords="offset points",
            alpha=0.8,
            bbox={"boxstyle": "round", "facecolor": "w"},
            wrap=True,
        )
        annot.set_visible(False)

        def hover_callback(event):
            if (
                event.inaxes is ax
                and event.xdata is not None
                and event.ydata is not None
            ):
                x, y = round(event.xdata), round(event.ydata)
                if self.df.values[y, x]:
                    annot.xy = (x, y)
                    frame = frames[x]
                    interaction = interactions[y]
                    residue = residues[y]
                    annot.set_text(f"Frame {frame}: {residue}")
                    color = im.cmap(self.color_mapper[interaction])
                    annot.get_bbox_patch().set_facecolor(color)
                    annot.set_visible(True)
                    fig.canvas.draw_idle()
                    return
            if annot.get_visible():
                annot.set_visible(False)
                fig.canvas.draw_idle()

        fig.canvas.mpl_connect("motion_notify_event", hover_callback)
        fig.canvas.header_visible = False
        fig.canvas.footer_visible = False



from functools import partial
from io import StringIO

import MDAnalysis as mda
import pytest

import prolif as plf
from prolif.exceptions import RunRequiredError
from prolif.plotting.network import LigNetwork


class TestLigNetwork:
    @pytest.fixture(scope="class")
    def simple_fp(self):
        return plf.Fingerprint(count=False)

    @pytest.fixture(scope="class")
    def count_fp(self):
        return plf.Fingerprint(count=True)

    @pytest.fixture(scope="class", params=["simple_fp", "count_fp"])
    def fp(self, request):
        return request.getfixturevalue(request.param)

    @pytest.fixture(scope="class")
    def fp_mol(self, fp):
        u = mda.Universe(plf.datafiles.TOP, plf.datafiles.TRAJ)
        lig = u.select_atoms("resname LIG")
        prot = u.select_atoms("protein and byres around 6.5 group ligand", ligand=lig)
        fp.run(u.trajectory[0:2], lig, prot)
        lig_mol = plf.Molecule.from_mda(lig)
        return fp, lig_mol

    @pytest.fixture(scope="class")
    def get_ligplot(self, fp_mol):
        fp, lig_mol = fp_mol
        return partial(LigNetwork.from_fingerprint, fp, lig_mol)

    def test_integration_frame(self, fp_mol):
        fp, lig_mol = fp_mol
        net = LigNetwork.from_fingerprint(
            fp, lig_mol, kind="frame", frame=0, display_all=fp.count
        )
        with StringIO() as buffer:
            net.save(buffer)
            buffer.seek(0)
            html = buffer.read()
        assert '"from": 18, "to": "PHE351.B", "title": "Hydrophobic' in html
        if fp.count:
            assert '"from": 19, "to": "PHE351.B", "title": "Hydrophobic' in html

    def test_integration_agg(self, get_ligplot):
        net = get_ligplot(kind="aggregate", threshold=0)
        with StringIO() as buffer:
            net.save(buffer)
            buffer.seek(0)
            html = buffer.read()
        assert "PHE331.B" in html

    def test_kwargs(self, get_ligplot):
        net = get_ligplot(
            kekulize=True,
            use_coordinates=True,
            flatten_coordinates=False,
            rotation=42,
            carbon=0,
        )
        with StringIO() as buffer:
            net.save(buffer)
            buffer.seek(0)
            html = buffer.read()
        assert "PHE331.B" in html

    def test_save_file(self, get_ligplot, tmp_path):
        net = get_ligplot()
        output = tmp_path / "lignetwork.html"
        net.save(output)
        with open(output, "r") as f:
            assert "PHE331.B" in f.read()

    def test_from_fingerprint_raises_kind(self, get_ligplot):
        with pytest.raises(ValueError, match='must be "aggregate" or "frame"'):
            get_ligplot(kind="foo")

    def test_from_fingerprint_raises_not_executed(self, ligand_mol):
        fp = plf.Fingerprint()
        with pytest.raises(
            RunRequiredError,
            match="Please run the fingerprint analysis before attempting to display results",
        ):
            LigNetwork.from_fingerprint(fp, ligand_mol)

    def test_fp_plot_lignetwork(self, fp_mol):
        fp, lig_mol = fp_mol
        html = fp.plot_lignetwork(lig_mol, kind="frame", frame=0, display_all=fp.count)
        assert "<iframe" in html.data



"""
Base interaction classes --- :mod:`prolif.interactions.base`
============================================================

This module contains the base classes used to build most of the interactions.
"""
import warnings
from itertools import product
from math import degrees, radians

import numpy as np
from rdkit import Geometry
from rdkit.Chem import MolFromSmarts

from prolif.interactions.utils import DISTANCE_FUNCTIONS, get_mapindex
from prolif.utils import angle_between_limits, get_centroid, get_ring_normal_vector

_INTERACTIONS = {}
_BASE_INTERACTIONS = {}


class Interaction:
    """Base class for interactions

    All interaction classes must inherit this class and define a :meth:`~detect` method.

    .. versionchanged:: 2.0.0
        Changed the return type of interactions. Added some helper methods to easily
        update/derive interaction classes.
    """

    def __init_subclass__(cls, is_abstract=False):
        super().__init_subclass__()
        name = cls.__name__
        register = _BASE_INTERACTIONS if is_abstract else _INTERACTIONS
        if not hasattr(cls, "detect"):
            raise TypeError(
                f"Can't instantiate interaction class {name} without a `detect` method."
            )
        if name in register:
            warnings.warn(
                f"The {name!r} interaction has been superseded by a "
                f"new class with id {id(cls):#x}"
            )
        register[name] = cls

    def __call__(self, lig_res, prot_res, metadata=False):
        for int_data in self.detect(lig_res, prot_res):
            yield int_data if metadata else True

    def __repr__(self):  # pragma: no cover
        cls = self.__class__
        return f"<{cls.__module__}.{cls.__name__} at {id(self):#x}>"

    def metadata(self, lig_res, prot_res, lig_indices, prot_indices, **data):
        """Returns a dict containing the indices of atoms responsible for the
        interaction, alongside any other metrics (e.g. distance, angle...etc.)."""
        if hasattr(self, "_metadata_mapping"):
            mapping = self._metadata_mapping or {}
            data = {mapping.get(key, key): value for key, value in data.items()}
        return {
            "indices": {
                "ligand": lig_indices,
                "protein": prot_indices,
            },
            "parent_indices": {
                "ligand": tuple(
                    [get_mapindex(lig_res, index) for index in lig_indices]
                ),
                "protein": tuple(
                    [get_mapindex(prot_res, index) for index in prot_indices]
                ),
            },
            **data,
        }

    @staticmethod
    def _invert_metadata(metadata):
        """Invert the role of the ligand and protein components in the dict returned
        by :meth:`~Interaction.metadata`."""
        if metadata is not None:
            metadata["indices"]["protein"], metadata["indices"]["ligand"] = (
                metadata["indices"]["ligand"],
                metadata["indices"]["protein"],
            )
            (
                metadata["parent_indices"]["protein"],
                metadata["parent_indices"]["ligand"],
            ) = (
                metadata["parent_indices"]["ligand"],
                metadata["parent_indices"]["protein"],
            )
        return metadata

    @classmethod
    def invert_role(cls, name, docstring):
        """Creates a new interaction class where the role of the ligand and protein
        residues have been swapped. Usefull to create e.g. an acceptor class from a
        donor class.
        """
        cls_docstring = cls.__doc__ or "\n"
        parameters_doc = cls_docstring.split("\n", maxsplit=1)[1]
        __doc__ = f"{docstring}\n{parameters_doc}"
        inverted = type(name, (cls,), {"__doc__": __doc__})

        def detect(self, ligand, residue):
            for metadata in super(inverted, self).detect(residue, ligand):
                yield self._invert_metadata(metadata)

        inverted.detect = detect
        return inverted


class Distance(Interaction, is_abstract=True):
    """Generic class for distance-based interactions

    Parameters
    ----------
    lig_pattern : str
        SMARTS pattern for atoms in ligand residues
    prot_pattern : str
        SMARTS pattern for atoms in protein residues
    distance : float
        Cutoff distance, measured between the first atom of each pattern
    """

    def __init__(self, lig_pattern, prot_pattern, distance):
        self.lig_pattern = MolFromSmarts(lig_pattern)
        self.prot_pattern = MolFromSmarts(prot_pattern)
        self.distance = distance

    def detect(self, lig_res, prot_res):
        lig_matches = lig_res.GetSubstructMatches(self.lig_pattern)
        prot_matches = prot_res.GetSubstructMatches(self.prot_pattern)
        if lig_matches and prot_matches:
            for lig_match, prot_match in product(lig_matches, prot_matches):
                alig = Geometry.Point3D(*lig_res.xyz[lig_match[0]])
                aprot = Geometry.Point3D(*prot_res.xyz[prot_match[0]])
                dist = alig.Distance(aprot)
                if dist <= self.distance:
                    yield self.metadata(
                        lig_res, prot_res, lig_match, prot_match, distance=dist
                    )


class SingleAngle(Interaction, is_abstract=True):
    """Generic class for interactions using constraints on a distance and an angle.

    Parameters
    ----------
    lig_pattern : str
        SMARTS pattern for atoms in ligand residues.
    prot_pattern : str
        SMARTS pattern for atoms in protein residues, should include (at least) two
        atoms, the first two atoms will be used for the angle calculation.
    distance : float
        Distance threshold.
    angle : tuple
        Min and max values for the ``[P1]-[P2]...[L1]`` angle, where L1 is the first
        atom in the ``lig_pattern`` SMARTS, and P1 and P2 are the first two atoms in the
        ``prot_pattern`` SMARTS.
    distance_atom : str
        Which atom from the protein (P1 or P2) to use for the distance calculation with
        L1.
    metadata_mapping : dict, optional
        Mapping for names used in the metadata dict for the distance and angle variables


    .. versionadded:: 2.0.0

    """

    def __init__(
        self,
        lig_pattern,
        prot_pattern,
        distance,
        angle,
        distance_atom,
        metadata_mapping=None,
    ):
        self.lig_pattern = MolFromSmarts(lig_pattern)
        self.prot_pattern = MolFromSmarts(prot_pattern)
        self.distance = distance
        self.angle = tuple(radians(i) for i in angle)
        self._measure_distance = DISTANCE_FUNCTIONS[distance_atom]
        self._metadata_mapping = metadata_mapping

    def detect(self, lig_res, prot_res):
        lig_matches = lig_res.GetSubstructMatches(self.lig_pattern)
        prot_matches = prot_res.GetSubstructMatches(self.prot_pattern)
        if lig_matches and prot_matches:
            for lig_match, prot_match in product(lig_matches, prot_matches):
                l1 = Geometry.Point3D(*lig_res.xyz[lig_match[0]])
                p1 = Geometry.Point3D(*prot_res.xyz[prot_match[0]])
                p2 = Geometry.Point3D(*prot_res.xyz[prot_match[1]])
                dist = self._measure_distance(l1, p1, p2)
                if dist <= self.distance:
                    # P1-P2 ... L1
                    p2p1 = p2.DirectionVector(p1)
                    p2l1 = p2.DirectionVector(l1)
                    angle = p2p1.AngleTo(p2l1)
                    if angle_between_limits(angle, *self.angle):
                        yield self.metadata(
                            lig_res,
                            prot_res,
                            lig_match,
                            prot_match,
                            distance=dist,
                            angle=degrees(angle),
                        )


class DoubleAngle(Interaction, is_abstract=True):
    """Generic class for interactions using constraints on a distance and two angles.

    Parameters
    ----------
    lig_pattern : str
        SMARTS for ``[L1]-[L2]``
    prot_pattern : str
        SMARTS for ``[P1]-[P2]``
    distance : float
        Distance threshold
    L1P2P1_angle : tuple
        Min and max values for the ``[L1]...[P2]-[P1]`` angle
    L2L1P2_angle : tuple
        Min and max values for the ``[L2]-[L1]...[P2]`` angle
    distance_atoms : tuple[str, str]
        Which atoms to use for the distance calculation: L1 or L2, and P1 or P2
    metadata_mapping : dict, optional
        Mapping for names used in the metadata dict for the distance and angle variables


    .. versionadded:: 2.0.0

    """

    def __init__(
        self,
        lig_pattern,
        prot_pattern,
        distance,
        L1P2P1_angle,
        L2L1P2_angle,
        distance_atoms=("L1", "P2"),
        metadata_mapping=None,
    ):
        self.lig_pattern = MolFromSmarts(lig_pattern)
        self.prot_pattern = MolFromSmarts(prot_pattern)
        self.distance = distance
        self.L1P2P1_angle = tuple(radians(i) for i in L1P2P1_angle)
        self.L2L1P2_angle = tuple(radians(i) for i in L2L1P2_angle)
        self._measure_distance = DISTANCE_FUNCTIONS[distance_atoms]
        self._metadata_mapping = metadata_mapping

    def detect(self, lig_res, prot_res):
        lig_matches = lig_res.GetSubstructMatches(self.lig_pattern)
        prot_matches = prot_res.GetSubstructMatches(self.prot_pattern)
        if lig_matches and prot_matches:
            for lig_match, prot_match in product(lig_matches, prot_matches):
                l1 = Geometry.Point3D(*lig_res.xyz[lig_match[0]])
                l2 = Geometry.Point3D(*lig_res.xyz[lig_match[1]])
                p1 = Geometry.Point3D(*prot_res.xyz[prot_match[0]])
                p2 = Geometry.Point3D(*prot_res.xyz[prot_match[1]])
                dist = self._measure_distance(l1, l2, p1, p2)
                if dist <= self.distance:
                    p2p1 = p2.DirectionVector(p1)
                    p2l1 = p2.DirectionVector(l1)
                    l1p2p1 = p2p1.AngleTo(p2l1)
                    if angle_between_limits(l1p2p1, *self.L1P2P1_angle):
                        l1p2 = l1.DirectionVector(p2)
                        l1l2 = l1.DirectionVector(l2)
                        l2l1p2 = l1p2.AngleTo(l1l2)
                        if angle_between_limits(l2l1p2, *self.L2L1P2_angle):
                            yield self.metadata(
                                lig_res,
                                prot_res,
                                lig_match,
                                prot_match,
                                distance=dist,
                                L1P2P1_angle=degrees(l1p2p1),
                                L2L1P2_angle=degrees(l2l1p2),
                            )


class BasePiStacking(Interaction, is_abstract=True):
    """Base class for Pi-Stacking interactions

    Parameters
    ----------
    distance : float
        Cutoff distance between each rings centroid
    plane_angles : tuple
        Min and max values for the angle between the ring planes
    normal_centroids_angles : tuple
        Min and max angles allowed between the vector normal to a ring's plane,
        and the vector between the centroid of both rings.
    pi_ring : list
        List of SMARTS for aromatic rings
    intersect : bool
        Whether to look for the point of intersection between both rings (for T-shaped
        interaction).
    intersect_radius : float
        If ``intersect=True``, used to check whether the intersect point falls within
        ``intersect_radius`` of the opposite ring's centroid.


    .. versionchanged:: 1.1.0
        The implementation now relies on the angle between the vector normal to a ring's
        plane and the vector between centroids (``normal_centroids_angles``) instead of
        the ``shortest_distance`` parameter.

    .. versionchanged:: 2.0.0
        Renamed ``centroid_distance`` to distance. Added ``intersect`` and
        ``ring_radius`` parameters.

    """

    def __init__(
        self,
        distance,
        plane_angle,
        normal_to_centroid_angle,
        pi_ring=(
            "[a;r6]1:[a;r6]:[a;r6]:[a;r6]:[a;r6]:[a;r6]:1",
            "[a;r5]1:[a;r5]:[a;r5]:[a;r5]:[a;r5]:1",
        ),
        intersect=False,
        intersect_radius=1.5,
    ):
        self.pi_ring = [MolFromSmarts(s) for s in pi_ring]
        self.distance = distance
        self.plane_angle = tuple(radians(i) for i in plane_angle)
        self.normal_to_centroid_angle = tuple(
            radians(i) for i in normal_to_centroid_angle
        )
        self.intersect = intersect
        self.intersect_radius = intersect_radius

    def detect(self, ligand, residue):
        for pi_rings in product(self.pi_ring, repeat=2):
            res_matches = residue.GetSubstructMatches(pi_rings[0])
            lig_matches = ligand.GetSubstructMatches(pi_rings[1])
            if not (lig_matches and res_matches):
                continue
            for lig_match, res_match in product(lig_matches, res_matches):
                lig_pi_coords = ligand.xyz[list(lig_match)]
                lig_centroid = Geometry.Point3D(*get_centroid(lig_pi_coords))
                res_pi_coords = residue.xyz[list(res_match)]
                res_centroid = Geometry.Point3D(*get_centroid(res_pi_coords))
                centroid_dist = lig_centroid.Distance(res_centroid)
                if centroid_dist > self.distance:
                    continue
                # ligand
                lig_normal = get_ring_normal_vector(lig_centroid, lig_pi_coords)
                # residue
                res_normal = get_ring_normal_vector(res_centroid, res_pi_coords)
                plane_angle = lig_normal.AngleTo(res_normal)
                if not angle_between_limits(plane_angle, *self.plane_angle, ring=True):
                    continue
                c1c2 = lig_centroid.DirectionVector(res_centroid)
                c2c1 = res_centroid.DirectionVector(lig_centroid)
                n1c1c2 = lig_normal.AngleTo(c1c2)
                n2c2c1 = res_normal.AngleTo(c2c1)
                ncc_angle = None
                if angle_between_limits(
                    n1c1c2, *self.normal_to_centroid_angle, ring=True
                ):
                    ncc_angle = n1c1c2
                elif angle_between_limits(
                    n2c2c1, *self.normal_to_centroid_angle, ring=True
                ):
                    ncc_angle = n2c2c1
                if ncc_angle is None:
                    continue
                if self.intersect:
                    # look for point of intersection between both ring planes
                    intersect = self._get_intersect_point(
                        lig_normal, lig_centroid, res_normal, res_centroid
                    )
                    if intersect is None:
                        continue
                    # check if intersection point falls ~within plane ring
                    intersect_dist = min(
                        lig_centroid.Distance(intersect),
                        res_centroid.Distance(intersect),
                    )
                    if intersect_dist <= self.intersect_radius:
                        yield self.metadata(
                            ligand,
                            residue,
                            lig_match,
                            res_match,
                            distance=centroid_dist,
                            plane_angle=degrees(plane_angle),
                            normal_to_centroid_angle=degrees(ncc_angle),
                            intersect_distance=intersect_dist,
                        )
                else:
                    yield self.metadata(
                        ligand,
                        residue,
                        lig_match,
                        res_match,
                        distance=centroid_dist,
                        plane_angle=degrees(plane_angle),
                        normal_to_centroid_angle=degrees(ncc_angle),
                    )

    @staticmethod
    def _get_intersect_point(
        plane_normal,
        plane_centroid,
        tilted_normal,
        tilted_centroid,
    ):
        # intersect line is orthogonal to both planes normal vectors
        intersect_direction = plane_normal.CrossProduct(tilted_normal)
        # setup system of linear equations to solve
        A = np.array(
            [list(plane_normal), list(tilted_normal), list(intersect_direction)]
        )
        if np.linalg.det(A) == 0:
            return None
        tilted_offset = tilted_normal.DotProduct(Geometry.Point3D(*tilted_centroid))
        plane_offset = plane_normal.DotProduct(Geometry.Point3D(*plane_centroid))
        d = np.array([[plane_offset], [tilted_offset], [0.0]])
        # point on intersect line
        point = np.linalg.solve(A, d).T[0]
        point = Geometry.Point3D(*point)
        # find projection of centroid on intersect line using vector projection
        vec = plane_centroid - point
        intersect_direction.Normalize()
        scalar_proj = intersect_direction.DotProduct(vec)
        return point + intersect_direction * scalar_proj



from copy import deepcopy
from math import radians

import numpy as np
import pytest
from numpy.testing import assert_equal
from rdkit import Chem

from prolif.residue import Residue, ResidueGroup, ResidueId
from prolif.utils import (
    angle_between_limits,
    get_centroid,
    get_residues_near_ligand,
    is_peptide_bond,
    pandas_series_to_bv,
    split_mol_by_residues,
    to_bitvectors,
    to_dataframe,
)


@pytest.fixture
def ifp_single():
    return {
        0: {
            ("LIG", "ALA1"): {"A": ({"indices": {"ligand": (0,), "protein": (1,)}},)},
            ("LIG", "GLU2"): {"B": ({"indices": {"ligand": (1,), "protein": (3,)}},)},
        },
        1: {
            ("LIG", "ALA1"): {
                "A": ({"indices": {"ligand": (2,), "protein": (4,)}},),
                "B": ({"indices": {"ligand": (2,), "protein": (5,)}},),
            },
            ("LIG", "ASP3"): {"B": ({"indices": {"ligand": (8,), "protein": (10,)}},)},
        },
    }


@pytest.fixture
def ifp_count():
    return {
        0: {
            ("LIG", "ALA1"): {
                "A": (
                    {"indices": {"ligand": (0,), "protein": (1,)}},
                    {"indices": {"ligand": (1,), "protein": (1,)}},
                )
            },
            ("LIG", "GLU2"): {"B": ({"indices": {"ligand": (1,), "protein": (3,)}},)},
        },
        1: {
            ("LIG", "ALA1"): {
                "A": (
                    {"indices": {"ligand": (2,), "protein": (4,)}},
                    {"indices": {"ligand": (2,), "protein": (1,)}},
                ),
                "B": ({"indices": {"ligand": (2,), "protein": (5,)}},),
            },
            ("LIG", "ASP3"): {"B": ({"indices": {"ligand": (8,), "protein": (10,)}},)},
        },
    }


@pytest.fixture(params=["ifp_single", "ifp_count"])
def ifp(request):
    return request.getfixturevalue(request.param)


def test_centroid():
    xyz = np.array(
        [(0, 0, 0), (0, 0, 0), (0, 0, 0), (2, 2, 2), (2, 2, 2), (2, 2, 2)],
        dtype=np.float32,
    )
    ctd = get_centroid(xyz)
    assert ctd.shape == (3,)
    assert_equal(ctd, [1, 1, 1])


@pytest.mark.parametrize(
    "angle, mina, maxa, ring, expected",
    [
        (0, 0, 30, False, True),
        (30, 0, 30, False, True),
        (10, 0, 30, False, True),
        (60, 0, 30, False, False),
        (150, 0, 30, False, False),
        (150, 0, 30, True, True),
        (60, 0, 30, True, False),
        (120, 0, 30, True, False),
    ],
)
def test_angle_limits(angle, mina, maxa, ring, expected):
    angle = radians(angle)
    mina = radians(mina)
    maxa = radians(maxa)
    assert angle_between_limits(angle, mina, maxa, ring) is expected


def test_pocket_residues(ligand_mol, protein_mol):
    resids = get_residues_near_ligand(ligand_mol, protein_mol)
    residues = [
        "TYR38.A",
        "TYR40.A",
        "GLN41.A",
        "VAL102.A",
        "SER106.A",
        "TYR109.A",
        "THR110.A",
        "TRP115.A",
        "TRP125.A",
        "LEU126.A",
        "ASP129.A",
        "ILE130.A",
        "THR131.A",
        "CYS133.A",
        "THR134.A",
        "ILE137.A",
        "ILE180.A",
        "GLU198.A",
        "CYS199.A",
        "VAL200.A",
        "VAL201.A",
        "ASN202.A",
        "THR203.A",
        "TYR208.A",
        "THR209.A",
        "VAL210.A",
        "TYR211.A",
        "SER212.A",
        "THR213.A",
        "VAL214.A",
        "GLY215.A",
        "ALA216.A",
        "PHE217.A",
        "TRP327.B",
        "PHE330.B",
        "PHE331.B",
        "ILE333.B",
        "SER334.B",
        "LEU335.B",
        "MET337.B",
        "PRO338.B",
        "LEU348.B",
        "ALA349.B",
        "ILE350.B",
        "PHE351.B",
        "ASP352.B",
        "PHE353.B",
        "PHE354.B",
        "THR355.B",
        "TRP356.B",
        "GLY358.B",
        "TYR359.B",
    ]
    for res in residues:
        r = ResidueId.from_string(res)
        assert r in resids


def test_split_residues():
    sequence = "ARNDCQEGHILKMFPSTWYV"
    prot = Chem.MolFromSequence(sequence)
    rg = ResidueGroup(
        [Residue(res) for res in Chem.SplitMolByPDBResidues(prot).values()]
    )
    residues = [Residue(mol) for mol in split_mol_by_residues(prot)]
    residues.sort(key=lambda x: x.resid)
    for molres, res in zip(residues, rg.values()):
        assert molres.resid == res.resid
        assert molres.HasSubstructMatch(res) and res.HasSubstructMatch(molres)


def test_is_peptide_bond():
    mol = Chem.RWMol()
    for i in range(3):
        a = Chem.Atom(6)
        mol.AddAtom(a)
    mol.AddBond(0, 1)
    mol.AddBond(1, 2)
    resids = {
        0: "ALA1.A",
        1: "ALA1.A",
        2: "GLU2.A",
    }
    assert is_peptide_bond(mol.GetBondWithIdx(0), resids) is False
    assert is_peptide_bond(mol.GetBondWithIdx(1), resids) is True


def test_series_to_bv():
    v = np.array([0, 1, 1, 0, 1])
    bv = pandas_series_to_bv(v)
    assert bv.GetNumBits() == len(v)
    assert bv.GetNumOnBits() == 3


def test_to_df(ifp):
    df = to_dataframe(ifp, ["A", "B", "C"])
    assert df.shape == (2, 4)
    assert df.dtypes[0].type is np.bool_
    assert df.index.name == "Frame"
    assert ("LIG", "ALA1", "A") in df.columns
    assert df[("LIG", "ALA1", "A")][0] is np.bool_(True)
    assert ("LIG", "ALA1", "B") in df.columns
    assert df[("LIG", "ALA1", "B")][0] is np.bool_(False)
    assert ("LIG", "ALA1", "C") not in df.columns
    assert ("LIG", "GLU2", "A") not in df.columns
    assert ("LIG", "ASP3", "B") in df.columns
    assert df[("LIG", "ASP3", "B")][0] is np.bool_(False)


@pytest.mark.parametrize(
    "dtype",
    [
        np.uint8,
        np.int16,
        np.bool_,
    ],
)
def test_to_df_dtype(dtype, ifp):
    df = to_dataframe(ifp, ["A", "B", "C"], dtype=dtype)
    assert df.dtypes[0].type is dtype
    assert df[("LIG", "ALA1", "A")][0] == dtype(True)
    assert df[("LIG", "ALA1", "B")][0] == dtype(False)
    assert df[("LIG", "ASP3", "B")][0] == dtype(False)


def test_to_df_drop_empty(ifp):
    df = to_dataframe(ifp, ["A", "B", "C"], drop_empty=False)
    assert df.shape == (2, 9)


def test_to_df_no_interaction_in_first_frame(ifp_single):
    fp = deepcopy(ifp_single)
    fp[0] = {}
    to_dataframe(fp, ["A", "B", "C"])


def test_to_df_count(ifp_count):
    df = to_dataframe(ifp_count, ["A", "B", "C"], count=True)
    assert df[df > 1].any().any()
    value = df[("LIG", "ALA1", "A")][0]
    assert value.dtype == np.uint8
    assert value == 2


def test_to_df_empty_ifp():
    ifp = {0: {}, 1: {}}
    df = to_dataframe(ifp, ["A"])
    assert df.to_numpy().shape == (2, 0)


def test_to_bv(ifp):
    df = to_dataframe(ifp, ["A", "B", "C"])
    bvs = to_bitvectors(df)
    assert len(bvs) == 2
    assert bvs[0].GetNumOnBits() == 2



"""
Residue-related classes --- :mod:`prolif.residue`
=================================================
"""
import re
from collections import UserDict
from typing import List, Optional

import numpy as np
from rdkit.Chem.rdmolops import FastFindRings

from prolif.rdkitmol import BaseRDKitMol

_RE_RESID = re.compile(r"([A-Z]{,3})?(\d*)\.?(\w)?")


class ResidueId:
    """A unique residue identifier

    Parameters
    ----------
    name : str
        3-letter residue name
    number : int
        residue number
    chain : str or None, optionnal
        1-letter protein chain
    """

    def __init__(self, name: str = "UNK", number: int = 0, chain: Optional[str] = None):
        self.name = name or "UNK"
        self.number = number or 0
        self.chain = chain or None

    def __repr__(self):
        return f"ResidueId({self.name}, {self.number}, {self.chain})"

    def __str__(self):
        resid = f"{self.name}{self.number}"
        if self.chain:
            resid += f".{self.chain}"
        return resid

    def __hash__(self):
        return hash((self.name, self.number, self.chain))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __lt__(self, other):
        return (self.chain, self.number) < (other.chain, other.number)

    @classmethod
    def from_atom(cls, atom):
        """Creates a ResidueId from an RDKit atom

        Parameters
        ----------
        atom : rdkit.Chem.rdchem.Atom
            An atom that contains an RDKit :class:`~rdkit.Chem.rdchem.AtomMonomerInfo`
        """
        mi = atom.GetMonomerInfo()
        if mi:
            name = mi.GetResidueName()
            number = mi.GetResidueNumber()
            chain = mi.GetChainId()
            return cls(name, number, chain)
        return cls()

    @classmethod
    def from_string(cls, resid_str):
        """Creates a ResidueId from a string

        Parameters
        ----------
        resid_str : str
            A string in the format ``<3-letter code><residue number>.<chain>``
            All arguments are optionnal, and the dot should be present only if
            the chain identifier is also present

        Examples
        --------

        +-----------+----------------------------------+
        | string    | Corresponding ResidueId          |
        +===========+==================================+
        | "ALA10.A" | ``ResidueId("ALA", 10, "A")``    |
        +-----------+----------------------------------+
        | "GLU33"   | ``ResidueId("GLU", 33, None)``   |
        +-----------+----------------------------------+
        | "LYS.B"   | ``ResidueId("LYS", 0, "B")``     |
        +-----------+----------------------------------+
        | "ARG"     | ``ResidueId("ARG", 0, None)``    |
        +-----------+----------------------------------+
        | "5.C"     | ``ResidueId("UNK", 5, "C")``     |
        +-----------+----------------------------------+
        | "42"      | ``ResidueId("UNK", 42, None)``   |
        +-----------+----------------------------------+
        | ".D"      | ``ResidueId("UNK", 0, "D")``     |
        +-----------+----------------------------------+
        | ""        | ``ResidueId("UNK", 0, None)``    |
        +-----------+----------------------------------+

        """
        matches = _RE_RESID.search(resid_str)
        name, number, chain = matches.groups()
        number = int(number) if number else 0
        return cls(name, number, chain)


class Residue(BaseRDKitMol):
    """A class for residues as RDKit molecules

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        The residue as an RDKit molecule

    Attributes
    ----------
    resid : prolif.residue.ResidueId
        The residue identifier

    Notes
    -----
    The name of the residue can be converted to a string by using
    ``str(Residue)``
    """

    def __init__(self, mol):
        super().__init__(mol)
        FastFindRings(self)
        self.resid = ResidueId.from_atom(self.GetAtomWithIdx(0))

    def __repr__(self):  # pragma: no cover
        name = ".".join([self.__class__.__module__, self.__class__.__name__])
        return f"<{name} {self.resid} at {id(self):#x}>"

    def __str__(self):
        return str(self.resid)


class ResidueGroup(UserDict):
    """A container to store and retrieve Residue instances easily

    Parameters
    ----------
    residues : list
        A list of :class:`~prolif.residue.Residue`

    Attributes
    ----------
    n_residues : int
        Number of residues in the ResidueGroup

    Notes
    -----
    Residues in the group can be accessed by :class:`ResidueId`, string, or
    index. See the :class:`~prolif.molecule.Molecule` class for an example.
    You can also use the :meth:`~prolif.residue.ResidueGroup.select` method to
    access a subset of a ResidueGroup.
    """

    def __init__(self, residues: List[Residue]):
        self._residues = np.asarray(residues, dtype=object)
        resinfo = [
            (r.resid.name, r.resid.number, r.resid.chain) for r in self._residues
        ]
        try:
            name, number, chain = zip(*resinfo)
        except ValueError:
            self.name = np.array([], dtype=object)
            self.number = np.array([], dtype=np.uint8)
            self.chain = np.array([], dtype=object)
        else:
            self.name = np.asarray(name, dtype=object)
            self.number = np.asarray(number, dtype=np.uint16)
            self.chain = np.asarray(chain, dtype=object)
        super().__init__([(r.resid, r) for r in self._residues])

    def __getitem__(self, key):
        # bool is a subclass of int but shouldn't be used here
        if isinstance(key, bool):
            raise KeyError(
                "Expected a ResidueId, int, or str, "
                f"got {type(key).__name__!r} instead"
            )
        if isinstance(key, int):
            return self._residues[key]
        elif isinstance(key, str):
            key = ResidueId.from_string(key)
            return self.data[key]
        elif isinstance(key, ResidueId):
            return self.data[key]
        raise KeyError(
            "Expected a ResidueId, int, or str, " f"got {type(key).__name__!r} instead"
        )

    def select(self, mask):
        """Locate a subset of a ResidueGroup based on a boolean mask

        Parameters
        ----------
        mask : numpy.ndarray
            A 1D array of ``dtype=bool`` with the same length as the number of
            residues in the ResidueGroup. The mask should be constructed by
            using conditions on the "name", "number", and "chain" residue
            attributes as defined in the :class:`~prolif.residue.ResidueId`
            class

        Returns
        -------
        rg : prolif.residue.ResidueGroup
            A subset of the original ResidueGroup

        Examples
        --------
        ::

            >>> rg
            <prolif.residue.ResidueGroup with 200 residues at 0x7f9a68719ac0>
            >>> rg.select(rg.chain == "A")
            <prolif.residue.ResidueGroup with 42 residues at 0x7fe3fdb86ca0>
            >>> rg.select((10 <= rg.number) & (rg.number < 30))
            <prolif.residue.ResidueGroup with 20 residues at 0x7f5f3c69aaf0>
            >>> rg.select((rg.chain == "B") & (np.isin(rg.name, ["ASP", "GLU"])))
            <prolif.residue.ResidueGroup with 3 residues at 0x7f5f3c510c70>

        As seen in these examples, you can combine masks with different
        operators, similarly to numpy boolean indexing or pandas
        :meth:`~pandas.DataFrame.loc` method

            * AND --> ``&``
            * OR --> ``|``
            * XOR --> ``^``
            * NOT --> ``~``

        """
        return ResidueGroup(self._residues[mask])

    def __repr__(self):  # pragma: no cover
        name = ".".join([self.__class__.__module__, self.__class__.__name__])
        return f"<{name} with {self.n_residues} residues at {id(self):#x}>"

    @property
    def n_residues(self):
        return len(self)



"""
Helper functions --- :mod:`prolif.utils`
========================================
"""
import warnings
from collections import defaultdict
from contextlib import contextmanager
from copy import deepcopy
from functools import wraps
from importlib.util import find_spec
from math import pi

import numpy as np
import pandas as pd
from rdkit import rdBase
from rdkit.Chem import FragmentOnBonds, GetMolFrags, SplitMolByPDBResidues
from rdkit.DataStructs import ExplicitBitVect, UIntSparseIntVect
from rdkit.Geometry import Point3D
from scipy.spatial import cKDTree

from prolif.residue import ResidueId

_90_deg_to_rad = pi / 2


def requires(module):  # pragma: no cover
    def inner(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if find_spec(module):
                return func(*args, **kwargs)
            raise ModuleNotFoundError(
                f"The module {module!r} is required to use {func.__name__!r} "
                "but it is not installed!"
            )

        return wrapper

    return inner


@contextmanager
def catch_rdkit_logs():
    log_status = rdBase.LogStatus()
    rdBase.DisableLog("rdApp.*")
    yield
    log_status = {st.split(":")[0]: st.split(":")[1] for st in log_status.split("\n")}
    log_status = {k: True if v == "enabled" else False for k, v in log_status.items()}
    for k, v in log_status.items():
        if v is True:
            rdBase.EnableLog(k)
        else:
            rdBase.DisableLog(k)


@contextmanager
def catch_warning(**kwargs):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", **kwargs)
        yield


def get_centroid(coordinates):
    """Centroid for an array of XYZ coordinates"""
    return np.mean(coordinates, axis=0)


def get_ring_normal_vector(centroid, coordinates):
    """Returns a vector that is normal to the ring plane"""
    # A & B are two edges of the ring
    a = Point3D(*coordinates[0])
    b = Point3D(*coordinates[1])
    # vectors between centroid and these edges
    ca = centroid.DirectionVector(a)
    cb = centroid.DirectionVector(b)
    # cross product between these two vectors
    normal = ca.CrossProduct(cb)
    # cb.CrossProduct(ca) is the normal vector in the opposite direction
    return normal


def angle_between_limits(angle, min_angle, max_angle, ring=False):
    """Checks if an angle value is between min and max angles in radian

    Parameters
    ----------
    angle : float
        Angle to check, in radians
    min_angle : float
        Lower bound angle, in radians
    max_angle : float
        Upper bound angle, in radians
    ring : bool
        Whether the angle being checked involves a ring or not

    Notes
    -----
    When ``ring=True``, the angle is capped between 0 and 90, and so should be
    the min and max angles. This is useful for angles involving a ring's plane
    normal vector.
    """
    if ring:
        if angle >= pi:
            angle %= _90_deg_to_rad
        elif angle > _90_deg_to_rad:
            angle = _90_deg_to_rad - (angle % _90_deg_to_rad)
    return min_angle <= angle <= max_angle


def get_residues_near_ligand(lig, prot, cutoff=6.0):
    """Detects residues close to a reference ligand

    Parameters
    ----------
    lig : prolif.molecule.Molecule or prolif.residue.Residue
        Select residues that are near this ligand/residue
    prot : prolif.molecule.Molecule
        Protein containing the residues
    cutoff : float
        If any interatomic distance between the ligand reference points and a
        residue is below or equal to this cutoff, the residue will be selected

    Returns
    -------
    residues : list
        A list of unique :class:`~prolif.residue.ResidueId` that are close to
        the ligand
    """
    tree = cKDTree(prot.xyz)
    ix = tree.query_ball_point(lig.xyz, cutoff)
    ix = set([i for lst in ix for i in lst])
    resids = [ResidueId.from_atom(prot.GetAtomWithIdx(i)) for i in ix]
    return list(set(resids))


def split_mol_by_residues(mol):
    """Splits a molecule in multiple fragments based on residues

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        The molecule to fragment

    Returns
    -------
    residues : list
        A list of :class:`rdkit.Chem.rdchem.Mol`

    Notes
    -----
    Code adapted from Maciek Wójcikowski on the RDKit discussion list
    """
    residues = []
    for res in SplitMolByPDBResidues(mol).values():
        for frag in GetMolFrags(res, asMols=True, sanitizeFrags=False):
            # count number of unique residues in the fragment
            resids = {a.GetIdx(): ResidueId.from_atom(a) for a in frag.GetAtoms()}
            if len(set(resids.values())) > 1:
                # split on peptide bonds
                bonds = [
                    b.GetIdx() for b in frag.GetBonds() if is_peptide_bond(b, resids)
                ]
                mols = FragmentOnBonds(frag, bonds, addDummies=False)
                mols = GetMolFrags(mols, asMols=True, sanitizeFrags=False)
                residues.extend(mols)
            else:
                residues.append(frag)
    return residues


def is_peptide_bond(bond, resids):
    """Checks if a bond is a peptide bond based on the ResidueId of the atoms
    on each part of the bond. Also works for disulfide bridges or any bond that
    links two residues in biopolymers.

    Parameters
    ----------
    bond : rdkit.Chem.rdchem.Bond
        The bond to check
    resids : dict
        A dictionnary of ResidueId indexed by atom index
    """
    return resids[bond.GetBeginAtomIdx()] != resids[bond.GetEndAtomIdx()]


def to_dataframe(
    ifp,
    interactions,
    count=False,
    dtype=None,
    drop_empty=True,
    index_col="Frame",
):
    """Converts IFPs to a pandas DataFrame

    Parameters
    ----------
    ifp : dict
        A dict in the format
        ``{<frame number>: {(<residue_id>, <residue_id>): <interactions>}}``.
        ``<interactions>`` is either a :class:`numpy.ndarray` bitvector, or a tuple of
        dict in the format ``{<interaction name>: <metadata dict>}``.
    interactions : list
        A list of interactions, in the same order as used to detect the interactions.
    count : bool
        Whether to output a count fingerprint or not.
    dtype : object or None
        Cast the dataframe values to this type. If ``None``, uses ``np.uint8`` if
        ``count=True``, else ``bool``.
    drop_empty : bool
        Drop columns with only empty values
    index_col : str
        Name of the index column in the DataFrame

    Returns
    -------
    df : pandas.DataFrame
        A 3-levels DataFrame where each ligand residue, protein residue, and
        interaction type are in separate columns

    Example
    -------
    ::

        >>> df = prolif.to_dataframe(results, fp.interactions.keys(), dtype=int)
        >>> print(df)
        ligand             LIG1.G
        protein             ILE59                  ILE55       TYR93
        interaction   Hydrophobic HBAcceptor Hydrophobic Hydrophobic PiStacking
        Frame
        0                       0          1           0           0          0
        ...

    .. versionchanged:: 0.3.2
        Moved the ``return_atoms`` parameter from the ``run`` methods to the
        dataframe conversion code

    .. versionchanged:: 2.0.0
        Removed the ``return_atoms`` parameter. Added the ``count`` parameter. Removed
        support for ``ifp`` containing ``np.ndarray`` bitvectors.
    """
    ifp = deepcopy(ifp)
    n_interactions = len(interactions)
    if dtype is None:
        dtype = np.uint8 if count else bool
    empty_value = dtype(0)
    # create empty array for each residue pair interaction that doesn't exist
    # in a particular frame
    empty_arr = np.array([empty_value for _ in range(n_interactions)], dtype=dtype)
    # residue pairs
    residue_pairs = sorted(
        set(
            [residue_tuple for frame_ifp in ifp.values() for residue_tuple in frame_ifp]
        )
    )
    # sparse to dense
    data = defaultdict(list)
    index = []
    for i, frame_ifp in ifp.items():
        index.append(i)
        for residue_tuple in residue_pairs:
            try:
                ifp_dict = frame_ifp[residue_tuple]
            except KeyError:
                data[residue_tuple].append(empty_arr[:])
            else:
                if count:
                    bitvector = np.array(
                        [len(ifp_dict.get(i, ())) for i in interactions], dtype=dtype
                    )
                else:
                    bitvector = np.array(
                        [i in ifp_dict for i in interactions], dtype=bool
                    )
                data[residue_tuple].append(bitvector)
    index = pd.Series(index, name=index_col)
    # create dataframe
    if not data:
        warnings.warn("No interaction detected")
        return pd.DataFrame([], index=index)
    values = np.array(
        [
            np.hstack([bitvector_list[frame] for bitvector_list in data.values()])
            for frame in range(len(index))
        ]
    )
    columns = pd.MultiIndex.from_tuples(
        [
            (str(lig_res), str(prot_res), i)
            for lig_res, prot_res in residue_pairs
            for i in interactions
        ],
        names=["ligand", "protein", "interaction"],
    )
    df = pd.DataFrame(values, columns=columns, index=index)
    df = df.astype(dtype)
    if drop_empty:
        mask = (df != empty_value).any(axis=0)
        df = df.loc[:, mask]
    return df


def pandas_series_to_bv(s):
    bv = ExplicitBitVect(len(s))
    on_bits = np.where(s >= True)[0].tolist()
    bv.SetBitsFromList(on_bits)
    return bv


def to_bitvectors(df):
    """Converts an interaction DataFrame to a list of RDKit ExplicitBitVector

    Parameters
    ----------
    df : pandas.DataFrame
        A DataFrame where each column corresponds to an interaction between two
        residues

    Returns
    -------
    bv : list
        A list of :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect`
        for each frame

    Example
    -------
    ::

        >>> from rdkit.DataStructs import TanimotoSimilarity
        >>> bv = prolif.to_bitvectors(df)
        >>> TanimotoSimilarity(bv[0], bv[1])
        0.42

    """
    return df.apply(pandas_series_to_bv, axis=1).tolist()


def pandas_series_to_countvector(s):
    size = len(s)
    cv = UIntSparseIntVect(size)
    for i in range(size):
        cv[i] = int(s[i])
    return cv


def to_countvectors(df):
    """Converts an interaction DataFrame to a list of RDKit UIntSparseIntVect

    Parameters
    ----------
    df : pandas.DataFrame
        A DataFrame where each column corresponds to the count for an interaction
        between two residues

    Returns
    -------
    cv : list
        A list of :class:`~rdkit.DataStructs.cDataStructs.UIntSparseIntVect`
        for each frame
    """
    return df.apply(pandas_series_to_countvector, axis=1).tolist()



"""
Reading RDKit molecules --- :mod:`prolif.rdkitmol`
==================================================
"""
from rdkit import Chem
from rdkit.Chem.rdMolTransforms import ComputeCentroid


class BaseRDKitMol(Chem.Mol):
    """Base molecular class that behaves like an RDKit :class:`~rdkit.Chem.rdchem.Mol`
    with extra attributes (see below).
    The sole purpose of this class is to define the common API between the
    :class:`~prolif.molecule.Molecule` and :class:`~prolif.residue.Residue` classes.
    This class should not be instantiated by users.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        A molecule (protein, ligand, or residue) with a single conformer

    Attributes
    ----------
    centroid : numpy.ndarray
        XYZ coordinates of the centroid of the molecule
    xyz : numpy.ndarray
        XYZ coordinates of all atoms in the molecule
    """

    @property
    def centroid(self):
        return ComputeCentroid(self.GetConformer())

    @property
    def xyz(self):
        return self.GetConformer().GetPositions()



import MDAnalysis as mda
import pytest

import prolif as plf
from prolif.exceptions import RunRequiredError
from prolif.plotting.complex3d import Complex3D


class TestComplex3D:
    @pytest.fixture(scope="class")
    def simple_fp(self):
        return plf.Fingerprint(count=False)

    @pytest.fixture(scope="class")
    def count_fp(self):
        return plf.Fingerprint(count=True)

    @pytest.fixture(scope="class", params=["simple_fp", "count_fp"])
    def fp(self, request):
        return request.getfixturevalue(request.param)

    @pytest.fixture(scope="class")
    def fp_mols(self, fp):
        u = mda.Universe(plf.datafiles.TOP, plf.datafiles.TRAJ)
        lig = u.select_atoms("resname LIG")
        prot = u.select_atoms("protein and byres around 6.5 group ligand", ligand=lig)
        fp.run(u.trajectory[0:2], lig, prot)
        lig_mol = plf.Molecule.from_mda(lig)
        prot_mol = plf.Molecule.from_mda(prot)
        return fp, lig_mol, prot_mol

    @pytest.fixture(scope="class")
    def plot_3d(self, fp_mols):
        fp, lig_mol, prot_mol = fp_mols
        return Complex3D.from_fingerprint(fp, lig_mol, prot_mol, frame=0)

    def test_integration_display_single(self, plot_3d):
        view = plot_3d.display(display_all=False)
        html = view._make_html()
        assert "Hydrophobic" in html

    def test_integration_display_all(self, plot_3d):
        view = plot_3d.display(display_all=True)
        html = view._make_html()
        assert "Hydrophobic" in html

    def test_from_fingerprint_raises_not_executed(self, ligand_mol, protein_mol):
        fp = plf.Fingerprint()
        with pytest.raises(
            RunRequiredError,
            match="Please run the fingerprint analysis before attempting to display results",
        ):
            Complex3D.from_fingerprint(
                fp,
                ligand_mol,
                protein_mol,
                frame=0,
            )

    def test_fp_plot_3d(self, fp_mols):
        fp, lig_mol, prot_mol = fp_mols
        view = fp.plot_3d(lig_mol, prot_mol, frame=0, display_all=fp.count)
        html = view._make_html()
        assert "Hydrophobic" in html



import MDAnalysis as mda
import pytest
from matplotlib import pyplot as plt

import prolif as plf
from prolif.exceptions import RunRequiredError
from prolif.plotting.barcode import Barcode


class TestBarcode:
    @pytest.fixture(scope="class")
    def simple_fp(self) -> plf.Fingerprint:
        return plf.Fingerprint(count=False)

    @pytest.fixture(scope="class")
    def count_fp(self) -> plf.Fingerprint:
        return plf.Fingerprint(count=True)

    @pytest.fixture(scope="class", params=["simple_fp", "count_fp"])
    def fp(self, request: pytest.FixtureRequest) -> plf.Fingerprint:
        return request.getfixturevalue(request.param)

    @pytest.fixture(scope="class")
    def fp_run(
        self, u: mda.Universe, ligand_ag, protein_ag, fp: plf.Fingerprint
    ) -> plf.Fingerprint:
        fp.run(u.trajectory[0:2], ligand_ag, protein_ag)
        return fp

    @pytest.fixture(scope="class")
    def barcode(self, fp_run: plf.Fingerprint) -> Barcode:
        return Barcode.from_fingerprint(fp_run)

    def test_display(self, barcode: Barcode) -> None:
        ax = barcode.display()
        assert isinstance(ax, plt.Axes)

    def test_display_kwargs(self, barcode: Barcode) -> None:
        ax = barcode.display(
            figsize=(1, 2),
            dpi=200,
            interactive=True,
            n_frame_ticks=2,
            residues_tick_location="bottom",
            xlabel="foobar",
            subplots_kwargs={},
            tight_layout_kwargs={"pad": 2},
        )
        assert isinstance(ax, plt.Axes)

    def test_fp_plot_barcode(self, fp_run: plf.Fingerprint) -> None:
        ax = fp_run.plot_barcode(
            figsize=(1, 2),
            dpi=200,
            interactive=True,
            n_frame_ticks=2,
            residues_tick_location="bottom",
            xlabel="foobar",
            subplots_kwargs={},
            tight_layout_kwargs={"pad": 2},
        )
        assert isinstance(ax, plt.Axes)

    def test_from_fingerprint_raises_not_executed(self) -> None:
        fp = plf.Fingerprint()
        with pytest.raises(
            RunRequiredError,
            match="Please run the fingerprint analysis before attempting to display results",
        ):
            Barcode.from_fingerprint(fp)



"""
Plot residues --- :mod:`prolif.plotting.residues`
=================================================

.. versionadded:: 2.0.0

.. autofunction:: display_residues

"""
from typing import Any, Optional, Tuple

from rdkit import Chem
from rdkit.Chem import Draw

from prolif.molecule import Molecule


def display_residues(
    mol: Molecule,
    residues_slice: Optional[slice] = None,
    *,
    size: Tuple[int, int] = (200, 140),
    mols_per_row: int = 4,
    use_svg: bool = True,
) -> Any:
    """Display a grid image of the residues in the molecule. The hydrogens are stripped
    and the 3D coordinates removed for a clearer visualisation.

    Parameters
    ----------
    mol: prolif.Molecule
        The molecule to show residues from.
    residues_slice: Optional[slice] = None
        Optionally, a slice of residues to display, e.g. ``slice(20)`` for the first 20
        residues, or ``slice(<start>, <stop>, <step>)`` for a more complex selection.
    size: Tuple[int, int] = (200, 140)
        Size of each residue image.
    mols_per_row: int = 4
        Number of residues displayed per row.
    use_svg: bool = True
        Generate an SVG or PNG image.
    """
    frags = []
    residues_iterable = (
        mol if residues_slice is None else mol.residues.select(residues_slice).values()
    )
    ipython_kwargs = (
        {"maxMols": mol.n_residues} if hasattr(Chem.Mol, "_repr_svg_") else {}
    )

    for residue in residues_iterable:
        resmol = Chem.RemoveHs(residue)
        resmol.RemoveAllConformers()
        resmol.SetProp("_Name", str(residue.resid))
        frags.append(resmol)

    return Draw.MolsToGridImage(
        frags,
        legends=[mol.GetProp("_Name") for mol in frags],
        subImgSize=size,
        molsPerRow=mols_per_row,
        useSVG=use_svg,
        **ipython_kwargs,
    )



"""
Generating an IFP in parallel --- :mod:`prolif.parallel`
========================================================

This module provides two classes, :class:`TrajectoryPool` and :class:`MolIterablePool`
to execute the analysis in parallel. These are used in the parallel implementation used
in :meth:`~prolif.fingerprint.Fingerprint.run` and
:meth:`~prolif.fingerprint.Fingerprint.run_from_iterable` respectively.
"""
from ctypes import c_uint32
from threading import Event, Thread
from time import sleep

from multiprocess import Value
from multiprocess.pool import Pool
from tqdm.auto import tqdm

from prolif.molecule import Molecule
from prolif.pickling import PICKLE_HANDLER


class Progress:
    """Helper class to track the number of frames processed by the
    :class:`TrajectoryPool`.

    Parameters
    ----------
    killswitch : threading.Event
        A threading.Event instance created and controlled by the :class:`TrajectoryPool`
        to kill the thread updating the :class:`~tqdm.std.tqdm` progress bar.
    tracker : multiprocess.Value
        Value holding a :class:`ctypes.c_uint32` ctype updated by the
        :class:`TrajectoryPool`, storing how many frames were processed since the last
        progress bar update.
    **kwargs : object
        Used to create the :class:`~tqdm.std.tqdm` progress bar.

    Attributes
    ----------
    delay : float, default = 0.5
        Delay between progress bar updates. This requires locking access to the
        ``tracker`` object which the :class:`TrajectoryPool` needs access to, so too
        small values might cause delays in the analysis.
    tqdm_pbar : tqdm.std.tqdm
        The progress bar displayed
    """

    delay = 0.5

    def __init__(self, killswitch, tracker, **kwargs):
        self.tqdm_pbar = tqdm(**kwargs)
        self.killswitch = killswitch
        self.tracker = tracker

    def close(self):
        """Cleanup after the :class:`TrajectoryPool` computation is done"""
        self.update()
        self.tqdm_pbar.close()

    def update(self):
        """Update the value displayed by the progress bar"""
        if self.tracker.value != 0:
            with self.tracker.get_lock():
                increment = self.tracker.value
                self.tracker.value = 0
            self.tqdm_pbar.update(increment)

    def event_loop(self):
        """Event loop targeted by a separate thread"""
        while True:
            if self.killswitch.is_set():
                break
            self.update()
            sleep(self.delay)


class TrajectoryPool:
    """Process pool for a parallelized IFP analysis on an MD trajectory. Must be used
    in a ``with`` statement.

    Parameters
    ----------
    n_processes : int
        Max number of processes
    fingerprint : prolif.fingerprint.Fingerprint
        Fingerprint instance used to generate the IFP
    residues : list, optional
        List of protein residues considered for the IFP
    tqdm_kwargs : dict
        Parameters for the :class:`~tqdm.std.tqdm` progress bar
    rdkitconverter_kwargs : tuple[dict, dict]
        Parameters for the :class:`~MDAnalysis.converters.RDKit.RDKitConverter`
        from MDAnalysis: the first for the ligand, and the second for the protein

    Attributes
    ----------
    tracker : multiprocess.Value
        Value holding a :class:`ctypes.c_uint32` ctype storing how many frames were
        processed since the last progress bar update.
    pool : multiprocess.pool.Pool
        The underlying pool instance.
    """

    def __init__(
        self, n_processes, fingerprint, residues, tqdm_kwargs, rdkitconverter_kwargs
    ):
        self.tqdm_kwargs = tqdm_kwargs
        self.tracker = Value(c_uint32, lock=True)
        self.pool = Pool(
            n_processes,
            initializer=self.initializer,
            initargs=(
                self.tracker,
                fingerprint,
                residues,
                rdkitconverter_kwargs,
            ),
        )

    @classmethod
    def initializer(cls, tracker, fingerprint, residues, rdkitconverter_kwargs):
        """Initializer classmethod passed to the pool so that each child process can
        access these objects without copying them."""
        cls.tracker = tracker
        cls.fp = fingerprint
        cls.residues = residues
        cls.converter_kwargs = rdkitconverter_kwargs

    @classmethod
    def executor(cls, args):
        """Classmethod executed by each child process on a chunk of the trajectory

        Returns
        -------
        ifp_chunk: dict[int, prolif.ifp.IFP]
            A dictionary of :class:`~prolif.ifp.IFP` indexed by frame number
        """
        traj, lig, prot, chunk = args
        ifp = {}
        for ts in traj[chunk]:
            lig_mol = Molecule.from_mda(lig, **cls.converter_kwargs[0])
            prot_mol = Molecule.from_mda(prot, **cls.converter_kwargs[1])
            data = cls.fp.generate(
                lig_mol, prot_mol, residues=cls.residues, metadata=True
            )
            ifp[int(ts.frame)] = data
            with cls.tracker.get_lock():
                cls.tracker.value += 1
        return ifp

    def process(self, args_iterable):
        """Maps the input iterable of arguments to the executor function.

        Parameters
        ----------
        args_iterable : typing.Iterable[tuple]
            Iterable of tuple of trajectory, ligand atomgroup, protein atomgroup, and
            array of frame indices.

        Returns
        -------
        ifp: typing.Iterable[dict[int, prolif.ifp.IFP]]
            An iterable of dictionaries of :class:`~prolif.ifp.IFP` indexed by frame
            number.
        """
        return self.pool.map(self.executor, args_iterable, chunksize=1)

    def __enter__(self):
        """Sets up the :class:`Progress` instance and associated killswitch event, and
        starts the progress event loop in a separate thread."""
        self.killswitch = Event()
        self.progress = Progress(self.killswitch, self.tracker, **self.tqdm_kwargs)
        self.progress_thread = Thread(target=self.progress.event_loop)
        self.progress_thread.start()
        return self

    def __exit__(self, *exc):
        """Call the killswitch and close the progress."""
        self.killswitch.set()
        self.progress.close()


class MolIterablePool:
    """Process pool for a parallelized IFP analysis on an iterable of ligands. Must be
    used in a ``with`` statement.

    Parameters
    ----------
    n_processes : int
        Max number of processes
    fingerprint : prolif.fingerprint.Fingerprint
        Fingerprint instance used to generate the IFP
    prot_mol : prolif.molecule.Molecule
        Protein molecule
    residues : list, optional
        List of protein residues considered for the IFP
    tqdm_kwargs : dict
        Parameters for the :class:`~tqdm.std.tqdm` progress bar

    Attributes
    ----------
    pool : multiprocess.pool.Pool
        The underlying pool instance.
    """

    def __init__(self, n_processes, fingerprint, prot_mol, residues, tqdm_kwargs):
        self.tqdm_kwargs = tqdm_kwargs
        self.pool = Pool(
            n_processes,
            initializer=self.initializer,
            initargs=(fingerprint, prot_mol, residues),
        )

    @classmethod
    def initializer(cls, fingerprint, prot_mol, residues):
        """Initializer classmethod passed to the pool so that each child process can
        access these objects without copying them."""
        cls.fp = fingerprint
        cls.pmol = prot_mol
        cls.residues = residues

    @classmethod
    def executor(cls, mol):
        """Classmethod executed by each child process on a single ligand molecule from
        the input iterable.

        Returns
        -------
        ifp_data : prolif.ifp.IFP
            A dictionary indexed by ``(ligand, protein)`` residue pairs, and each value
            is a sparse dictionary of metadata indexed by interaction name.
        """
        return cls.fp.generate(mol, cls.pmol, residues=cls.residues, metadata=True)

    def process(self, args_iterable):
        """Maps the input iterable of molecules to the executor function.

        Parameters
        ----------
        args_iterable : typing.Iterable[prolif.molecule.Molecule]
            An iterable yielding ligand molecules

        Returns
        -------
        ifp : typing.Iterable[prolif.ifp.IFP]
            An iterable of :class:`~prolif.ifp.IFP` dictionaries.
        """
        results = self.pool.imap(self.executor, args_iterable, chunksize=1)
        return tqdm(results, **self.tqdm_kwargs)

    def __enter__(self):
        """Sets up which properties will be pickled by RDKit by default"""
        PICKLE_HANDLER.set()
        return self

    def __exit__(self, *exc):
        """Resets RDKit's default pickled properties"""
        PICKLE_HANDLER.reset()



import pytest

from prolif.fingerprint import Fingerprint
from prolif.residue import ResidueId


@pytest.fixture(scope="session")
def ifp(u, ligand_ag, protein_ag):
    fp = Fingerprint(["Hydrophobic", "VdWContact"])
    fp.run(u.trajectory[0:1], ligand_ag, protein_ag)
    return fp.ifp[0]


def test_ifp_indexing(ifp):
    lig_id, prot_id = "LIG1.G", "LEU126.A"
    metadata1 = ifp[(ResidueId.from_string(lig_id), ResidueId.from_string(prot_id))]
    metadata2 = ifp[(lig_id, prot_id)]
    assert metadata1 is metadata2


def test_ifp_filtering(ifp):
    lig_id, prot_id = "LIG1.G", "LEU126.A"
    assert ifp[lig_id] == ifp
    assert (
        next(iter(ifp[prot_id].values()))
        == ifp[(ResidueId.from_string(lig_id), ResidueId.from_string(prot_id))]
    )


def test_wrong_key(ifp):
    with pytest.raises(KeyError, match="does not correspond to a valid IFP key"):
        ifp[0]



import pickle

import pytest
from rdkit import Chem

from prolif.pickling import PROLIF_PICKLE_OPTIONS, RDKitPickleHandler


@pytest.fixture(autouse=True)
def reset_default_pickle_properties():
    default = Chem.GetDefaultPickleProperties()
    yield
    Chem.SetDefaultPickleProperties(default)


def test_default():
    assert (
        RDKitPickleHandler(0).default_pickle
        == RDKitPickleHandler.get()
        == Chem.GetDefaultPickleProperties()
    )


def test_pickling():
    mol = Chem.MolFromSmiles("CCO")
    mol.SetProp("foo", "bar")
    atom = mol.GetAtomWithIdx(0)
    atom.SetProp("spam", "egg")
    assert mol.HasProp("foo") and atom.HasProp("spam")
    # should discard properties
    RDKitPickleHandler(Chem.PropertyPickleOptions.NoProps).set()
    pickled = pickle.loads(pickle.dumps(mol))
    assert not pickled.HasProp("foo")
    # should keep mol and atom properties
    handler = RDKitPickleHandler(PROLIF_PICKLE_OPTIONS)
    handler.set()
    pickled = pickle.loads(pickle.dumps(mol))
    assert pickled.GetProp("foo") == mol.GetProp("foo")
    assert pickled.GetAtomWithIdx(0).GetProp("spam") == atom.GetProp("spam")
    # should go back to previous pickle options (discard properties)
    handler.reset()
    pickled = pickle.loads(pickle.dumps(mol))
    assert not pickled.HasProp("foo")
